import requests
import pandas as pd
import json
from multiprocessing import Pool
import itertools
from .exceptions import NoMatchingDataError
from .parsers import parse_final_domain, parse_base_output, parse_monitoring
from typing import List, Dict
from .util import to_snake_case

__title__ = "jao-py"
__version__ = "0.5.2"
__author__ = "Frank Boerman"
__license__ = "MIT"


class JaoPublicationToolClient:
    BASEURL = "https://publicationtool.jao.eu/core/api/data/"

    def __init__(self, api_key: str = None):
        self.s = requests.Session()
        self.s.headers.update({
            'user-agent': f'jao-py {__version__} (github.com/fboerman/jao-py)'
        })

        if api_key is not None:
            self.s.headers.update({
                'Authorization': 'Bearer ' + api_key
            })

        self.NORDIC = 'nordic' in self.BASEURL

    def _starmap_pull(self, url, params, keyname=None):
        r = self.s.get(url, params=params)
        r.raise_for_status()
        if keyname is not None:
            return r.json()[keyname]
        else:
            return r.json()

    def _query_domain(self, url: str, mtu: pd.Timestamp, presolved: bool = None, cne: str = None, co: str = None,
                           urls_only: bool = False):
        if cne is not None or co is not None or bool is not None:
            filter = {
                'CnecName': "" if cne is None else cne,
                'Contingency': "" if co is None else co,
                'Presolved': presolved
            }
        else:
            filter = None

        # first do a call with zero retrieved data to know how much data is available, then pull all at once
        r = self.s.get(self.BASEURL + url, params={
            'FromUtc': mtu.isoformat(),
            'ToUtc': (mtu + pd.Timedelta(hours=1)).isoformat(),
            'Filter': json.dumps(filter),
            'Skip': 0,
            'Take': 0
        })
        r.raise_for_status()

        if r.json()['totalRowsWithFilter'] == 0:
            raise NoMatchingDataError

        # now do new call with all data requested
        # jao servers are not great returning it all at once, but they let you choose your own pagination
        # lets go for chunks of 5000, arbitrarily chosen

        total_num_data = r.json()['totalRowsWithFilter']
        args = []
        for i in range(0, total_num_data, 5000):
            args.append((self.BASEURL + url, {
                'FromUtc': mtu.isoformat(),
                'ToUtc': (mtu + pd.Timedelta(hours=1)).isoformat(),
                'Filter': json.dumps(filter),
                'Skip': i,
                'Take': 5000
            }, 'data'))

        if urls_only:
            return args

        with Pool() as pool:
            results = pool.starmap(self._starmap_pull, args)

        return list(itertools.chain(*results))

    def query_final_domain(self, mtu: pd.Timestamp, presolved: bool = None, cne: str = None, co: str = None,
                           urls_only: bool = False) -> List[Dict]:
        if type(mtu) != pd.Timestamp:
            raise Exception('Please use a timezoned pandas Timestamp object for mtu')
        if mtu.tzinfo is None:
            raise Exception('Please use a timezoned pandas Timestamp object for mtu')
        mtu = mtu.tz_convert('UTC')

        return self._query_domain('finalComputation', mtu=mtu, presolved=presolved, cne=cne, co=co, urls_only=urls_only)

    def _query_base_fromto(self, d_from: pd.Timestamp, d_to: pd.Timestamp, type: str) -> List[Dict]:
        if type in ['monitoring']:
            url = self.BASEURL.replace('/data/', '/system/')
        else:
            url = self.BASEURL
        r = self.s.get(url + type, params={
            'FromUTC': d_from.isoformat(),
            'ToUTC': d_to.isoformat()
        })
        r.raise_for_status()
        data = r.json()['data']
        if len(data) == 0:
            raise NoMatchingDataError
        return data

    def _query_base_day(self, day: pd.Timestamp, type: str) -> List[Dict]:
        d_from = day.replace(hour=0, minute=0)
        d_to = day.replace(hour=0, minute=0) + pd.Timedelta(days=1)
        # the api does some funny stuff on dst days, so adjust the to for this edge case
        if d_from.dst() > d_to.dst():
            d_to += pd.Timedelta(hours=1)
        elif d_from.dst() < d_to.dst():
            d_to -= pd.Timedelta(hours=1)
        return self._query_base_fromto(
            d_from=d_from,
            d_to=d_to,
            type=type
        )

    def query_net_position(self, day: pd.Timestamp) -> List[Dict]:
        return self._query_base_day(
            day=day,
            type='netPos'
        )

    def query_active_constraints(self, day: pd.Timestamp) -> List[Dict]:
        # although the same skip/take mechanism is active on this endpoint as the final domain, this is not needed to be used
        #   by definition active constraints are only a few so its overkill to start pagination
        # for the same reason this endpoint returns a whole day at once instead of per hour since there are not many
        #  and you probably want the whole day anyway
        # for the date range to be correct make sure the day input has a timezone!

        return self._query_base_day(
            day=day,
            type='shadowPrices'
        )

    def query_lta(self, d_from: pd.Timestamp, d_to: pd.Timestamp) -> List[Dict]:
        return self._query_base_fromto(d_from, d_to, 'lta')

    def query_validations(self, d_from: pd.Timestamp, d_to: pd.Timestamp) -> List[Dict]:
        return self._query_base_fromto(d_from, d_to, 'validationReductions')

    def query_maxbex(self, day: pd.Timestamp) -> List[Dict]:
        return self._query_base_day(day, 'maxExchanges')

    def query_minmax_np(self, day: pd.Timestamp) -> List[Dict]:
        return self._query_base_day(day, 'maxNetPos')

    def query_allocationconstraint(self, d_from: pd.Timestamp, d_to: pd.Timestamp) -> List[Dict]:
        return self._query_base_fromto(d_from, d_to, 'allocationConstraint')

    def query_status(self, d_from: pd.Timestamp, d_to: pd.Timestamp) -> List[Dict]:
        return self._query_base_fromto(d_from, d_to, 'spanningDefaultFBP')

    def query_price_spread(self, d_from: pd.Timestamp, d_to: pd.Timestamp) -> List[Dict]:
        return self._query_base_fromto(d_from, d_to, 'priceSpread')

    def query_scheduled_exchange(self, d_from: pd.Timestamp, d_to: pd.Timestamp) -> List[Dict]:
        return self._query_base_fromto(d_from, d_to, 'scheduledExchanges')

    def query_alpha_factor(self, d_from: pd.Timestamp, d_to: pd.Timestamp) -> List[Dict]:
        return self._query_base_fromto(d_from, d_to, 'alphaFactor')

    def query_monitoring(self, day: pd.Timestamp) -> List[Dict]:
        return self._query_base_day(day, 'monitoring')

    def query_d2cf(self, d_from: pd.Timestamp, d_to: pd.Timestamp) -> List[Dict]:
        return self._query_base_fromto(
            d_from=d_from, d_to=d_to,
            type='d2CF' if not self.NORDIC else 'cgmForeCast'
        )


class JaoPublicationToolPandasClient(JaoPublicationToolClient):
    def query_final_domain(self, mtu: pd.Timestamp, presolved: bool = None, cne: str = None,
                           co: str = None) -> pd.DataFrame:
        return parse_final_domain(
            super().query_final_domain(mtu=mtu, presolved=presolved, cne=cne, co=co)
        )

    def query_allocationconstraint(self, d_from: pd.Timestamp, d_to: pd.Timestamp) -> pd.DataFrame:
        return parse_base_output(
            super().query_allocationconstraint(d_from=d_from, d_to=d_to)
        ).rename(columns=lambda c: c.split('_')[1] + '_' + ('import' if 'Down' in c.split('_')[0] else 'export'))

    def query_net_position(self, day: pd.Timestamp) -> pd.DataFrame:
        return parse_base_output(
            super().query_net_position(day=day)
        ).rename(columns=lambda x: x.replace('hub_', '')) \
            .rename(columns={'DE': 'DE_LU'})

    def query_active_constraints(self, day: pd.Timestamp) -> pd.DataFrame:
        return parse_base_output(
            super().query_active_constraints(day=day)
        ).rename(columns=lambda x: to_snake_case(x) if 'hub' not in x else x) \
            .rename(columns={'id': 'id_original'}) \
            .rename(columns=lambda x: x.replace('hub_', 'ptdf_'))

    def query_maxbex(self, day: pd.Timestamp, from_zone: str = None, to_zone: str = None) -> pd.DataFrame:
        df = parse_base_output(
            super().query_maxbex(day=day)
        ).rename(columns=lambda x: x.lstrip('border_').replace('_', '>'))

        if from_zone is not None:
            df = df[[c for c in df.columns if c.split('>')[0] == from_zone]]

        if to_zone is not None:
            df = df[[c for c in df.columns if c.split('>')[1] == to_zone]]

        return df

    def query_minmax_np(self, day: pd.Timestamp) -> pd.DataFrame:
        return parse_base_output(
            super().query_minmax_np(day=day)
        )

    def query_lta(self, d_from: pd.Timestamp, d_to: pd.Timestamp) -> pd.DataFrame:
        return parse_base_output(
            super().query_lta(d_from=d_from, d_to=d_to)
        )

    def query_validations(self, d_from: pd.Timestamp, d_to: pd.Timestamp) -> pd.DataFrame:
        df = parse_base_output(
            super().query_validations(d_from=d_from, d_to=d_to)
        ).rename(columns=to_snake_case)
        return df

    def query_status(self, d_from: pd.Timestamp, d_to: pd.Timestamp) -> pd.DataFrame:
        return parse_base_output(
            super().query_status(d_from=d_from, d_to=d_to)
        ).drop(columns=['lastModifiedOn'])

    def query_alpha_factor(self, d_from: pd.Timestamp, d_to: pd.Timestamp) -> pd.DataFrame:
        return parse_base_output(
            super().query_alpha_factor(d_from=d_from, d_to=d_to)
        ).drop(columns=['lastModifiedOn'])

    def query_price_spread(self, d_from: pd.Timestamp, datetime, d_to: pd.Timestamp) -> (
            pd.DataFrame):
        return parse_base_output(
            super().query_price_spread(d_from=d_from, d_to=d_to)
        )

    def query_scheduled_exchange(self, d_from: pd.Timestamp, d_to: pd.Timestamp) -> (
            pd.DataFrame):
        return parse_base_output(
            super().query_scheduled_exchange(d_from=d_from, d_to=d_to)
        )

    def query_monitoring(self, day: pd.Timestamp) -> pd.DataFrame:
        return parse_monitoring(
            super().query_monitoring(day=day)
        )

    def query_d2cf(self, d_from: pd.Timestamp, d_to: pd.Timestamp) -> pd.DataFrame:
        return parse_base_output(
            super().query_d2cf(d_from=d_from, d_to=d_to)
        )


import pandas as pd
from lxml import etree
from datetime import datetime
from .definitions import ParseDataSubject
from typing import Union


def _infer_and_convert_type(s):
    try:
        v = s[s.first_valid_index()]
    except KeyError:
        # no valid index so whole thing is nan. simply return as float nan type
        return s.astype(float)

    if type(v) != str:
        return s
    t = None
    try:
        int(v)
        t = int
    except:
        try:
            float(v)
            t = float
        except:
            if v == 'true' or v == 'false':
                return s.map({'true': True, 'false': False}).astype(bool)
    if t is not None:
        try:
            return s.astype(t)
        except:
            pass
    return s.fillna('')


def _parse_utility_tool_xml(xml: Union[bytes, str], t: ParseDataSubject) -> pd.DataFrame:
    """
    parses the xml coming out of the utility tool

    :param xml: string or bytes of the xml data
    :param t: which type to parse, choose from "MaxExchanges", "MaxNetPositions", "Ptdfs"
    :return:
    """
    if type(xml) == str:
        xml_b = xml.encode('UTF-8')
    elif type(xml) == bytes:
        xml_b = xml
    else:
        raise ValueError("xml should be provided in either bytes or string format")

    tree = etree.fromstring(xml_b)
    # the node name is the singular of the type
    node_name = t.value.strip('s')
    # get all the column names by checking all the children tags of a node
    column_names = [x.tag for x in tree.xpath(t + '/' + node_name)[0].getchildren()]

    # now get all data fast with xpath by retrieving the text for every tag, the order is guaranteed the same
    # so we can stitch them back together later
    data = []

    for c_name in column_names:
        data.append([str(x) for x in tree.xpath(f'{t}/{node_name}/{c_name}/node()')])

    df = pd.DataFrame(zip(*data), columns=column_names)
    for c in column_names[2:]:
        df[c] = _infer_and_convert_type(df[c])

    df['TIMESTAMP_CET'] = df.apply(lambda row: datetime.strptime(row['Date'], '%Y-%m-%dT00:00:00').replace(hour=int(row['CalendarHour'])-1), axis=1)

    df.drop(columns=['Date', 'CalendarHour'], inplace=True)

    return df


def _parse_utilitytool_xml(xml: Union[bytes, str], nodename: str, columnnames: list, datenode: str,
                           xpath: str='') -> pd.DataFrame:
    """
    parses the xml coming out excell utilitytool endpoints which is xml

    :param xml: string or bytes of the xml data
    :return:
    """

    if type(xml) == str:
        xml_b = xml.encode('UTF-8')
    elif type(xml) == bytes:
        xml_b = xml
    else:
        raise ValueError("xml should be provided in either bytes or string format")

    tree = etree.fromstring(xml_b)

    data = {
        'date': tree.xpath(xpath + f"ns:{nodename}/ns:{datenode}/node()", namespaces={'ns': 'http://tempuri.org/'}),
        'hour': tree.xpath(xpath + f"ns:{nodename}/ns:CalendarHour/node()", namespaces={'ns': 'http://tempuri.org/'}),
    }
    for c in columnnames:
        data[c] = tree.xpath(xpath + f"ns:{nodename}/ns:{c}/node()", namespaces={'ns': 'http://tempuri.org/'})

    df = pd.DataFrame.from_dict(data, orient='columns')
    df = df[['date', 'hour'] + columnnames]
    df[columnnames] = df[columnnames].astype('float')
    df = pd.DataFrame(df)

    # TODO: this fails with DST, change to index pd.date_range instead per day
    df['timestamp'] = df.apply(
        lambda row: datetime.strptime(row['date'], '%Y-%m-%dT00:00:00').replace(hour=int(row['hour']) - 1),
        axis=1)
    df.drop(columns=['date', 'hour'], inplace=True)
    df = df.set_index('timestamp').tz_localize('europe/amsterdam')

    return df


def _parse_maczt_final_flowbased_domain(df: pd.DataFrame, zone='NL') -> pd.DataFrame:
    """
    extracts the MACZT numbers from the final flowbased domain dataframe
    for now only NL zone is supported

    :param df:
    :return:
    """
    if zone != 'NL':
        raise NotImplementedError

    # select only relevant columns
    df = df[['CO', 'CO_EIC', 'CNE', 'CNE_EIC',
             'Presolved', 'RAM', 'Fmax', 'Fref', 'AMR', 'MinRAMFactor', 'MinRAMFactorJustification']]

    # if there is a default parameter day there is an empty dataframe. stop further processing
    if len(df) == 0:
        return df


    # filter on cnecs that have a valid dutch justification string and are not lta
    # make sure to make copy to prevent slice errors later
    df = df[df['MinRAMFactorJustification'].str.contains('MACZTtarget').fillna(False)]
    df = df[~(df['CNE'].str.contains('LTA_corner'))]
    df = pd.DataFrame(df)

    df['MCCC_PCT'] = 100 * df['RAM'] / df['Fmax']

    df[['MNCC_PCT', 'LF_CALC_PCT', 'LF_ACCEPT_PCT', 'MACZT_TARGET_PCT']] = \
        df['MinRAMFactorJustification'].str.extract(
            r'MNCC = (?P<MNCC_PCT>.*)%;LFcalc = (?P<LF_CALC_PCT>.*)%;LFaccept = (?P<LF_ACCEPT_PCT>.*)%;MACZTtarget = (?P<MACZT_TARGET_PCT>.*)%')
    df[['MCCC_PCT', 'MNCC_PCT', 'LF_CALC_PCT', 'LF_ACCEPT_PCT', 'MACZT_TARGET_PCT']] = \
        df[['MCCC_PCT', 'MNCC_PCT', 'LF_CALC_PCT', 'LF_ACCEPT_PCT', 'MACZT_TARGET_PCT']].astype(float)

    df['MACZT_PCT'] = df['MCCC_PCT'] + df['MNCC_PCT']
    df['LF_SUB_PCT'] = (df['LF_CALC_PCT'] - df['LF_ACCEPT_PCT']).clip(lower=0)
    df['MACZT_MIN_PCT'] = df['MACZT_TARGET_PCT'] - df['LF_SUB_PCT']
    df['MACZT_MARGIN'] = df['MACZT_PCT'] - df['MACZT_MIN_PCT']

    df.drop(columns=['MinRAMFactorJustification'], inplace=True)

    # there is only two decimals statistical relevance here so round it at that
    df[['MCCC_PCT', 'MACZT_PCT', 'LF_SUB_PCT', 'MACZT_MIN_PCT', 'MACZT_MARGIN']] = df[['MCCC_PCT', 'MACZT_PCT', 'LF_SUB_PCT', 'MACZT_MIN_PCT', 'MACZT_MARGIN']].round(2)

    return df


def _parse_suds_tradingdata(data, subject: str, nested: bool = False, freq: str = 'H') -> pd.DataFrame:
    """

    :param data: suds.sudsobject.TradingData object
    :param subject: string of the type that needs to be parsed
    :param freq: timefrequency for
    :return: resulting pandas dataframe
    """
    data_parsed = []
    if nested:
        data_raw = data[subject][subject.strip('s')]
    else:
        data_raw = data[subject]

    date_column = 'Date' if 'Date' in data_raw[0] else 'CalendarDate'

    df = pd.DataFrame([dict(x) for x in data_raw])
    df.index = pd.date_range(df[date_column].min(),
                             df[date_column].max().replace(hour=23),
                             freq=freq, tz='Europe/Amsterdam')
    df.drop(columns=[date_column, 'CalendarHour'], inplace=True)
    return df.astype(float)



from .jao import JaoPublicationToolPandasClient
import pandas as pd
from .parsers import parse_final_domain

class JaoPublicationToolPandasNordics(JaoPublicationToolPandasClient):
    BASEURL = "https://publicationtool.jao.eu/nordic/api/data/"

    def query_allocationconstraint(self, d_from: pd.Timestamp, d_to: pd.Timestamp) -> pd.DataFrame:
        raise NotImplementedError

    def query_net_position(self, day: pd.Timestamp) -> pd.DataFrame:
        raise NotImplementedError

    def query_lta(self, d_from: pd.Timestamp, d_to: pd.Timestamp) -> pd.DataFrame:
        raise NotImplementedError

    def query_alpha_factor(self, d_from: pd.Timestamp, d_to: pd.Timestamp) -> pd.DataFrame:
        raise NotImplementedError

    def query_price_spread(self, d_from: pd.Timestamp, datetime, d_to: pd.Timestamp) -> (
            pd.DataFrame):
        raise NotImplementedError

    def query_scheduled_exchange(self, d_from: pd.Timestamp, d_to: pd.Timestamp) -> (
            pd.DataFrame):
        raise NotImplementedError

    def query_active_constraints(self, mtu: pd.Timestamp, shadow_price_only: bool = False) -> pd.DataFrame:
        df = parse_final_domain(
            super()._query_domain('fbDomainShadowPrice', mtu=mtu)
        )

        if shadow_price_only:
            df = pd.DataFrame(df[~pd.isna(df['shadow_price'])])

        return df


import requests
import pandas as pd
from datetime import timedelta, date
from calendar import monthrange
from dateutil.relativedelta import relativedelta
from typing import Union


class JaoAPIClient:
    # https://www.jao.eu/page-api/market-data

    BASEURL = "https://api.jao.eu/OWSMP/"

    def __init__(self, api_key):
        self.s = requests.Session()
        self.s.headers.update({
            'user-agent': 'jao-py (github.com/fboerman/jao-py)',
            'AUTH_API_KEY': api_key
        })

    def query_auction_corridors(self):
        r = self.s.get(self.BASEURL + 'getcorridors')
        r.raise_for_status()
        return [x['value'] for x in r.json()]

    def query_auction_horizons(self):
        r = self.s.get(self.BASEURL + 'gethorizons')
        r.raise_for_status()
        return [x['value'] for x in r.json()]

    def query_auction_details_by_month(self, corridor: str, month: date, horizon: str, shadow_auctions_only: bool = False) -> dict:
        """
        get the auction data for a specified month. gives basically everything but the bids themselves

        :param shadow_auctions_only: wether to only retrieve shadow auction results
        :param corridor: string of a valid jao corridor from query_corridors
        :param month: datetime.date object for the month you want the auction data of
        :param horizon: string object for the horizon you want the auction data of
        :return:
        """
        # prepare the specific input arguments needed, start day the day before the months begin
        # end date the last day of the month

        month_begin = month.replace(day=1)
        month_end = month.replace(day=monthrange(month.year, month.month)[1])
        if horizon == 'Yearly':
            r = self.s.get(self.BASEURL + 'getauctions', params={
                'corridor': corridor,
                'fromdate': month.strftime('%Y-%m-%d'),
                'horizon': horizon,
                'shadow': int(shadow_auctions_only)
            })
        else:
            r = self.s.get(self.BASEURL + 'getauctions', params={
                'corridor': corridor,
                'fromdate': (month_begin - timedelta(days=1)).strftime("%Y-%m-%d"),
                'horizon': horizon,
                'todate': month_end.strftime("%Y-%m-%d"),
                'shadow': int(shadow_auctions_only)
            })

        r.raise_for_status()

        data = r.json()
        # pretify the results since we know it is for monthly auction
        data = data[0]
        data = {**data, **data['results'][0], **data['products'][0]}
        del data['results']
        del data['products']

        return data

    def query_auction_bids_by_month(self, corridor: str, month: date, as_dict: bool = False) -> Union[pd.DataFrame, dict]:
        """
        wrapper function to construct the auction id since its predictable and
          pass it on to the query function for the bids

        :param corridor: string of a valid jao corridor from query_corridors
        :param month: datetime.date object for the month you want the auction bids of
        :param as_dict: boolean wether the result needs to be returned as a dictionary instead of a dataframe
        :return:
        """

        return self.query_auction_bids_by_id(month.strftime(f"{corridor}-M-BASE-------%y%m01-01"), as_dict=as_dict)

    def query_curtailments_by_month(self, corridor: str, month: date, as_dict: bool = False) -> Union[pd.DataFrame, list]:
        month_begin = month.replace(day=1)
        month_end = month.replace(day=monthrange(month.year, month.month)[1])
        r = self.s.get(self.BASEURL + 'getcurtailment', params={
            'corridor': corridor,
            'fromdate': (month_begin - timedelta(days=1)).strftime("%Y-%m-%d"),
            'todate': month_end.strftime("%Y-%m-%d"),
        })

        r.raise_for_status()
        if as_dict:
            return r.json()
        else:
            df = pd.DataFrame(r.json())
            df['curtailmentPeriodStart'] = pd.to_datetime(df['curtailmentPeriodStart'])\
                    .dt.tz_convert('europe/amsterdam')
            df['curtailmentPeriodStop'] = pd.to_datetime(df['curtailmentPeriodStop'])\
                    .dt.tz_convert('europe/amsterdam')
            return df

    def query_auction_bids_by_id(self, auction_id: str, as_dict: bool = False) -> Union[pd.DataFrame, list]:
        r = self.s.get(self.BASEURL + "getbids", params={
            'auctionid': auction_id
        })

        r.raise_for_status()

        if as_dict:
            return r.json()
        else:
            return pd.DataFrame(r.json())

    def query_auction_stats_months(self, month_from: date, month_to: date, corridor: str, horizon: str = 'Monthly') -> pd.DataFrame:
        """
        gets the following statistics for the give range of months (included both ends) in a dataframe:
        id
        corridor
        month
        auction start
        auction ended
        offered capacity (MW)
        ATC (MW)
        allocated capacity (MW)
        resold capacity (MW)
        requested capacity (MW)
        price (EUR/MW)
        non allocated capacity (MW)
        in this order

        :param month_from: datetime.date object of start month
        :param month_to: datetime.date object of end month
        :param corridor: string of a valid jao corridor from query_corridors
        :param horizon: string of a valid jao horizon from query_horizons
        :return:
        """
        detail_keys = ['bidGateOpening', 'bidGateClosure', 'offeredCapacity', 'atc',
                        'allocatedCapacity', 'resoldCapacity', 'requestedCapacity', 'auctionPrice']

        data = []
        m = month_from
        while m <= month_to:
            m_details = self.query_auction_details_by_month(corridor=corridor, month=m, horizon=horizon)
            m_data = {
                'id': m_details['identification'],
                'corridor': corridor,
                'month': m.replace(day=1)
            }
            m_data = {**m_data, **{k: v for k, v in m_details.items() if k in detail_keys}}
            data.append(m_data)
            m += relativedelta(months=1)
        df = pd.DataFrame(data)
        df['resoldCapacity'].fillna(0, inplace=True)
        df['nonAllocatedCapacity'] = df['offeredCapacity'] - df['allocatedCapacity']

        return df



from .jao import JaoPublicationToolClient, JaoPublicationToolPandasClient
from .jao_intraday import JaoPublicationToolPandasIntraDay
from .jao_parrun import JaoPublicationToolPandasIntraDayParRun
from .jao_nordic import JaoPublicationToolPandasNordics
from .webservice import JaoAPIClient



from .jao import JaoPublicationToolPandasClient
import pandas as pd
from .parsers import parse_base_output
from typing import List, Dict


class JaoPublicationToolPandasIntraDayParRun(JaoPublicationToolPandasClient):
    def __init__(self, version):
        super().__init__()
        if version == 'c':
            self.BASEURL = "https://parallelrun-publicationtool.jao.eu/coreID/api/data/IDCCC_"
        else:
            raise NotImplementedError

    def query_lta(self, d_from: pd.Timestamp, d_to: pd.Timestamp):
        raise NotImplementedError

    def query_status(self, d_from: pd.Timestamp, d_to: pd.Timestamp):
        raise NotImplementedError

    def query_active_constraints(self, day: pd.Timestamp):
        raise NotImplementedError

    def query_alpha_factor(self, d_from: pd.Timestamp, d_to: pd.Timestamp):
        raise NotImplementedError

    def query_monitoring(self, day: pd.Timestamp) -> List[Dict]:
        # use this quick hack because this monitoring endpoint differs from all others
        base_url_old = self.BASEURL
        self.BASEURL = self.BASEURL.replace("IDCCC_", '')
        data = self._query_base_day(day, 'monitoring')
        self.BASEURL = base_url_old
        return data

    def query_sidc_atc(self, day: pd.Timestamp, from_zone: str = None, to_zone: str = None) -> pd.DataFrame:
        df = parse_base_output(
            self._query_base_day(day, 'intradayAtc')
        ).rename(columns=lambda x: x.lstrip('border_').replace('_', '>'))

        if from_zone is not None:
            df = df[[c for c in df.columns if c.split('>')[0] == from_zone]]

        if to_zone is not None:
            df = df[[c for c in df.columns if c.split('>')[1] == to_zone]]

        return df

    def query_sidc_ntc(self, day: pd.Timestamp, from_zone: str = None, to_zone: str = None) -> pd.DataFrame:
        df = parse_base_output(
            self._query_base_day(day, 'intradayNtc')
        ).rename(columns=lambda x: x.lstrip('border_').replace('_', '>'))

        if from_zone is not None:
            df = df[[c for c in df.columns if c.split('>')[0] == from_zone]]

        if to_zone is not None:
            df = df[[c for c in df.columns if c.split('>')[1] == to_zone]]

        return df



import pandas as pd
from .util import to_snake_case
from typing import List, Dict


def parse_final_domain(data: List[Dict]) -> pd.DataFrame:
    # flatten the data
    # note: this only selects first CO
    # save the order of keys to keep output consistent
    # (note in p3.7+ dict order is guaranteed: https://stackoverflow.com/a/39980744

    columns = list(data[0].keys())
    # get keys of second level and insert them at right place in columns
    i = columns.index('contingencies')
    columns = columns[:i] + ['contingency_' + x for x in data[0]['contingencies'][0].keys()] + columns[i + 1:]
    for d in data:
        for c_k, c_d in d['contingencies'][0].items():
            d['contingency_' + c_k] = c_d
        del d['contingencies']
    # now build the dataframe and convert column names
    df = pd.DataFrame(data)
    df = df.rename(columns=lambda x: to_snake_case(x) if 'ptdf' not in x else x)
    # also convert our earlier column list
    columns = [to_snake_case(x) if 'ptdf' not in x else x for x in columns]
    # drop needless columns
    columns.remove('contingency_number')
    df = df.drop(columns=['contingency_number'])
    # fix order of columns
    df = df[columns]
    df = df.rename(columns={'id': 'id_original'})
    # parse datetime, convert to localtime and adjust column name
    df['date_time_utc'] = pd.to_datetime(df['date_time_utc'], utc=True).dt.tz_convert('europe/amsterdam')
    df = df.rename(columns={'date_time_utc': 'mtu'})
    # return the result!
    return df


def parse_monitoring(data: List[Dict]) -> pd.DataFrame:
    df = pd.DataFrame(data)
    if 'businessDayUtc' in df:
        df['businessDay'] = pd.to_datetime(df['businessDayUtc'], utc=True).dt.tz_convert('europe/amsterdam').dt.date
        df = df.drop(columns=['businessDayUtc'])
    for c in ['deadline', 'lastModifiedOn']:
        df[c] = pd.to_datetime(df[c], utc=True).dt.tz_convert('europe/amsterdam')
    df=df.drop(columns=['id'])
    return df


def parse_base_output(data: List[Dict]) -> pd.DataFrame:
    df = pd.DataFrame(data).drop(columns='id')
    df['dateTimeUtc'] = pd.to_datetime(df['dateTimeUtc'], utc=True).dt.tz_convert('europe/amsterdam')
    df = df.set_index('dateTimeUtc')
    df.index.name = 'mtu'
    return df


import requests
import pandas as pd
from .exceptions import ServerReturnedEmptyData, InvalidCaptcha
from datetime import date
from suds.client import Client as suds_Client
from functools import wraps
from PIL import Image
from io import BytesIO, StringIO
from .parsers import _parse_utility_tool_xml, _parse_maczt_final_flowbased_domain, \
    _parse_utilitytool_xml, _parse_suds_tradingdata
import numpy as np
from .definitions import ParseDataSubject


class JaoUtilityToolASMXClient:
    # from the ASMX Web Service API, this is a very good defined system
    #   which supplies the endpoint and formats in xml upfront. this is delegate to the suds package
    #   for convenience some methods are wrapped into a dataframe

    def __init__(self):
        # all communication goes through the sud package, since the schemas are defined in the WSDL
        self.client = suds_Client("http://utilitytool.jao.eu/CascUtilityWebService.asmx?WSDL")

    def help(self):
        """
        this prints the scheme as defined by the WSDL format and parsed by the suds package

        :return:
        """
        print(str(self.client))

    def query_minmax_NP(self, d_from: str, d_to: str) -> pd.DataFrame:
        """

        :param d_from: start date string that is accepted by pandas timestamp
        :param d_to: end date string that is accepted by pandas timestamp
        """
        d_from = pd.Timestamp(d_from).strftime("%Y-%m-%d")
        d_to = pd.Timestamp(d_to).strftime("%Y-%m-%d")
        return _parse_suds_tradingdata(
            self.client.service.GetTradingDataForAPeriod(d_from, d_to, False, True, False),
            'MaxNetPositions',
            True
        )

    def query_max_bex(self, d_from: str, d_to: str) -> pd.DataFrame:
        """

        :param d_from: start date string that is accepted by pandas timestamp
        :param d_to: end date string that is accepted by pandas timestamp
        """
        d_from = pd.Timestamp(d_from).strftime("%Y-%m-%d")
        d_to = pd.Timestamp(d_to).strftime("%Y-%m-%d")
        return _parse_suds_tradingdata(
            self.client.service.GetTradingDataForAPeriod(d_from, d_to, True, False, False),
            'MaxExchanges',
            True
        )

    def query_CWE_NP(self, d_from: str, d_to: str) -> pd.DataFrame:
        """

        :param d_from: start date string that is accepted by pandas timestamp
        :param d_to: end date string that is accepted by pandas timestamp
        """
        d_from = pd.Timestamp(d_from).strftime("%Y-%m-%d")
        d_to = pd.Timestamp(d_to).strftime("%Y-%m-%d")
        return _parse_suds_tradingdata(
            self.client.service.GetNetPositionDataForAPeriod(d_from, d_to),
            'NetPositionData'
        )


class JaoUtilityToolCSVClient:
    # this ingests from the same client the excel macros in the utility tool is talking too
    # because of this the endpoints are open and thus no key or captcha is required
    def __init__(self):
        self.s = requests.Session()
        self.s.headers.update({
            'user-agent': 'jao-py (github.com/fboerman/jao-py)'
        })

    def query_cwe_net_position(self, d_from: str, d_to: str) -> pd.DataFrame:
        """
        Downloads the internal cwe net positions between the given date range

        :param d_from: start date string that is accepted by pandas timestamp
        :param d_to: end date string that is accepted by pandas timestamp
        :return:
        """
        d_from = pd.Timestamp(d_from)
        d_to = pd.Timestamp(d_to)
        url = f"https://utilitytool.jao.eu/WebServiceV2.asmx/GetNetPositionDataForAPeriod?" \
              f"dateFrom={d_from.strftime('%m-%d-%Y')}&dateTo={d_to.strftime('%m-%d-%Y')}"
        r = self.s.get(url)
        r.raise_for_status()

        return _parse_utilitytool_xml(r.text, 'NetPositionData', ['AT', 'NL', 'BE', 'DE', 'FR', 'ALBE', 'ALDE'], 'CalendarDate')

    def query_cwe_minmax_NP(self, d_from: str, d_to: str) -> pd.DataFrame:
        d_from = pd.Timestamp(d_from)
        d_to = pd.Timestamp(d_to)
        url = f"https://utilitytool.jao.eu/WebServiceV2.asmx/GetTradingDataForAPeriod?" \
              f"dateFrom={d_from.strftime('%m-%d-%Y')}&dateTo={d_to.strftime('%m-%d-%Y')}" \
              f"&maxExchange=false&netPosition=true&ptdf=false"
        r = self.s.get(url)
        r.raise_for_status()

        return _parse_utilitytool_xml(r.text, 'MaxNetPosition',
                                      ['MinAT', 'MaxAT', 'MinNL', 'MaxNL', 'MinBE', 'MaxBE', 'MinDE', 'MaxDE', 'MinFR',
                                       'MaxFR', 'MinALBE', 'MaxALBE', 'MinALDE', 'MaxALDE'], 'Date', xpath='ns:MaxNetPositions/')

    def _parse_domain(self, r: requests.Response) -> pd.DataFrame:
        # do some character formatting so pandas understands the text
        lines = r.text.replace(';|', "|").split('\r\n')
        lines[0] = lines[0].replace(';', '|')
        # load it in a virtual file and give it to pandas
        stream = StringIO()
        stream.write("\n".join(lines))
        stream.seek(0)

        df = pd.read_csv(stream, sep="|")

        # check if the dataframe is empty, this should not happen. default flow parameters always return something.
        # throw an error and let the user deal with it
        if len(df) == 0:
            raise ServerReturnedEmptyData

        # parse the date string which is only the day
        df['DeliveryDate'] = pd.to_datetime(df['DeliveryDate'], format='%d/%m/%Y %H:%M:%S')
        # insert the hour into the date by combining with the Period column

        # for DST: difficult problem. JAO gives everything in localtime, so we cant use the easy trick of everything in UTC and then later convert
        # it also gives the hour only as a period number for that day
        # for now this package solves it the following way: throwing away the double hour in case of clock going backwards
        # and leaving the hour empty when clock is going forward.
        # If a reader knows a better way please open an issue or pullrequest on github
        if df['Period'].max() > 24:
            # clock going backwards so one extra hour at hour
            def _shift_hour(row):
                if row['Period'] < 4: # before clock change
                    return row['DeliveryDate'].replace(hour=row['Period'] - 1)
                elif row['Period'] == 4: # during clock change
                    return np.nan
                elif row['Period'] > 4: # after clock change
                    return row['DeliveryDate'].replace(hour=row['Period'] - 2)
            df['DeliveryDate'] = df.apply(_shift_hour, axis=1)
        elif df['Period'].max() < 24:
            # clock going forward so one hour less
            def _shift_hour(row):
                if row['Period'] < 3: # before clock change
                    return row['DeliveryDate'].replace(hour=row['Period'] - 1)
                elif row['Period'] >= 3: # after clock change
                    return row['DeliveryDate'].replace(hour=row['Period'])
            df['DeliveryDate'] = df.apply(_shift_hour, axis=1)
        else:
            # normal time
            df['DeliveryDate'] = df.apply(lambda row: row['DeliveryDate'].replace(hour=row['Period'] - 1), axis=1)
        df = df.dropna(subset=['DeliveryDate'])
        df = df.rename(columns={'DeliveryDate': 'timestamp'}).drop(columns=['Period']).set_index('timestamp')
        df = df.tz_localize('Europe/Amsterdam', ambiguous=True)

        # now do some cleanup, remove useless columns and make the ptdf columns more efficient
        df.drop(columns=['FileId', 'Row'], inplace=True)
        # for ptdf columns we assume the jao system that the same bidding zone is mentioned in whole column
        # so only check first entry
        ptdf_translation = {}
        useless_columns = []
        i = 0
        while True:
            if i == 0:
                c1 = "Factor"
                c2 = "BiddingArea_Shortname"
            else:
                c1 = f"Factor.{i}"
                c2 = f"BiddingArea_Shortname.{i}"
            if c1 in df.columns and c2 in df.columns:
                ptdf_translation[c1] = f"PTDF_{df[c2].iloc[0]}"
                useless_columns.append(c2)
            else:
                break
            i += 1
        df = df.rename(columns=ptdf_translation)
        df = df.drop(columns=useless_columns)
        df = df.rename(columns={
                'OutageName': 'CO',
                'OutageEIC': 'CO_EIC',
                'CriticalBranchName': 'CNE',
                'CriticalBranchEIC': 'CNE_EIC',
                'RemainingAvailableMargin': 'RAM'
            })

        return df

    def query_final_flowbased_domain(self, d: str) -> pd.DataFrame:
        """
        Downloads the final flowbased of the business day of the given date object
        returns a dataframe with the data. This endpoint is relatively slow so we download one day per request
        for multiple days call this function repeatedly and pd.concat the dataframes

        :param d: date string that is accepted by pandas timestamp
        """
        d = pd.Timestamp(d)

        # retrieve the data from jao network call
        url = f"https://utilitytool.jao.eu/CSV/GetAllCBCOFixedLabelDataForAPeriod?dateFrom={d.strftime('%m-%d-%Y')}&" \
              f"dateTo={d.strftime('%m-%d-%Y')}&random=1"
        r = self.s.get(url)
        # check for http errors
        r.raise_for_status()

        return self._parse_domain(r)

    def query_initial_virgin_domain(self, d: str) -> pd.DataFrame:
        """
        Works the same as query_final_flowbased_domain but for the virgin domain initial computation

        :param d: date string that is accepted by pandas timestamp
        """
        d = pd.Timestamp(d)

        # retrieve the data from jao network call
        url = f"https://utilitytool.jao.eu/CSV/GetVirginDomainInitialComputationDataForAPeriod?"\
              f"dateFrom={d.strftime('%m-%d-%Y')}&" \
              f"dateTo={d.strftime('%m-%d-%Y')}&random=1"
        r = self.s.get(url)
        # check for http errors
        r.raise_for_status()

        return self._parse_domain(r)

    def query_final_virgin_domain(self, d: str) -> pd.DataFrame:
        """
        Works the same as query_final_flowbased_domain but for the virgin domain final computation

        :param d: date string that is accepted by pandas timestamp
        """
        d = pd.Timestamp(d)

        # retrieve the data from jao network call
        url = f"https://utilitytool.jao.eu/CSV/GetVirginDomainFinalComputationDataForAPeriod?"\
              f"dateFrom={d.strftime('%m-%d-%Y')}&" \
              f"dateTo={d.strftime('%m-%d-%Y')}&random=1"
        r = self.s.get(url)
        # check for http errors
        r.raise_for_status()

        return self._parse_domain(r)

    def query_maczt(self, d: str, zone: str = 'NL') -> pd.DataFrame:
        """
        Extract the MACZT numbers from the final flowbased domain.
        Calls the query_final_flowbased_domain function and has thus the same limitation about
                being relatively slow and only one day per call
        For now only the NL zone is supported

        :param d: date string that is accepted by pandas timestamp
        :param zone: str of the selected zone
        :return:
        """

        if zone != 'NL':
            raise NotImplementedError

        return _parse_maczt_final_flowbased_domain(
            self.query_final_flowbased_domain(d),
            zone=zone
        )


def captcha(func):
    """
    check if captcha is already solved and otherwise request one and solve it

    :param func:
    :return:
    """
    @wraps(func)
    def captcha_wrapper(*args, **kwargs):
        self = args[0]

        # atm the only check is if we have ever done the captcha
        #  later add checks if it has expired (does it even expire?)
        if self.captcha is None:
            r = self.s.get(self.BASEURL + "/Captcha/Show")
            r.raise_for_status()
            png_stream = BytesIO(r.content)
            png_stream.seek(0)
            captcha_png = Image.open(png_stream)
            captcha_png.show()
            captcha_code = input("Captcha: ")
            r = self.s.get(self.BASEURL + "/Util/Validate", params={
                'captchaValue': captcha_code
            })
            if r.status_code != 200:
                raise InvalidCaptcha
            if r.text != 'True':
                raise InvalidCaptcha
            self.captcha = captcha_code

        return func(*args, **kwargs)

    return captcha_wrapper


class JaoUtilityToolXmlClient:
    # uses the xml download from the utility tool website at https://www.jao.eu/implict-allocation
    # this requires solving a recaptcha by the user

    BASEURL = "https://utilitytool.jao.eu"

    def __init__(self):
        self.s = requests.Session()
        self.s.headers.update({
            'user-agent': 'jao-py (github.com/fboerman/jao-py)'
        })
        self.captcha = None

    @captcha
    def query_xml(self, date_from: str, date_to: str) -> bytes:
        """
        download the utility tool xml file with given dates, returns the raw xml in bytes

        :param date_from: start date string that is accepted by pandas timestamp
        :param date_to: end date string that is accepted by pandas timestamp
        :return:
        """

        date_from = pd.Timestamp(date_from)
        date_to = pd.Timestamp(date_to)

        r = self.s.post(self.BASEURL + '/Util/Download', data={
            'Date': date.today().strftime("%Y/%m/%d"),
            'fileType': 'Xml',
            'FromDate': date_from.strftime("%Y/%m/%d"),
            'ToDate': date_to.strftime("%Y/%m/%d"),
            'CaptchaValue': self.captcha,
            'InvisibleCaptchaValue': '',
            'force': 'false',
        })

        r.raise_for_status()

        return r.content

    def query_df(self, date_from: str, date_to: str, t: ParseDataSubject) -> pd.DataFrame:
        """
        downloads the utility tool xml file and parses the selected data type from the response into a dataframe

        :param date_from: start date string that is accepted by pandas timestamp
        :param date_to: end date string that is accepted by pandas timestamp
        :param t: which type to parse, choose from ParseDataSubject Enum
        :return:
        """

        return _parse_utility_tool_xml(self.query_xml(date_from, date_to), t)



# -*- coding: utf-8 -*-

"""
A setuptools based setup module for jao-py

Adapted from
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Get the version from the source code
with open(path.join(here, 'jao', 'jao.py'), encoding='utf-8') as f:
    lines = f.readlines()
    for l in lines:
        if l.startswith('__version__'):
            __version__ = l.split('"')[1]

setup(
    name='jao-py',
    version=__version__,
    description='A python API wrapper for JAO.eu',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/fboerman/jao-py',
    author='Frank Boerman',
    author_email='frank@fboerman.nl',
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',

        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],

    keywords='JAO data api energy',

    packages=find_packages(),

    install_requires=['requests', 'pandas', 'python-dateutil', 'beautifulsoup4'],

    package_data={
        'jao-py': ['LICENSE.md', 'README.md'],
    },
)
